﻿using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace TraceSpy
{
    public class ColorRange : IEquatable<ColorRange>, IComparable, IComparable<ColorRange>
    {
        internal ColorRange(ColorSet colorset, int startIndex, int length)
        {
            ColorSet = colorset;
            StartIndex = startIndex;
            Length = length;
        }

        public int StartIndex { get; private set; }
        public int Length { get; private set; }
        public ColorSet ColorSet { get; private set; }

        internal static IEnumerable<ColorRange> ComputeColorRanges(Settings settings, string line)
        {
            if (settings == null)
                throw new ArgumentNullException("settings");

            Colorizer[] colorizers = settings.Colorizers;
            if (string.IsNullOrEmpty(line))
            {
                yield return new ColorRange(null, 0, 0);
                yield break;
            }

            if ((colorizers == null) || (colorizers.Length == 0) ||
                (settings.ColorSets == null) || (settings.ColorSets.Length == 0))
            {
                yield return new ColorRange(null, 0, line.Length);
                yield break;
            }

            // get named groups that match one of the defined color set
            Main.Log("");
            Main.Log("Compute line(" + line.Length + "):'" + line + "'");
            List<ColorRange> ranges = new List<ColorRange>();
            foreach (Colorizer colorizer in colorizers)
            {
                if (!colorizer.Active)
                    continue;

                Main.Log(" Match colorizer:" + colorizer.Definition);
                foreach (Match m in colorizer.Regex.Matches(line))
                {
                    for (int i = 0; i < m.Groups.Count; i++)
                    {
                        string name = colorizer.Regex.GroupNameFromNumber(i);
                        if (string.IsNullOrEmpty(name))
                            continue;

                        Group group = m.Groups[i];
                        if (!group.Success)
                            continue;

                        ColorSet set = settings.GetColorSet(name);
                        if (set == null)
                            continue;

                        Main.Log("  Match group name:" + name + " value=" + group.Value);
                        ColorRange range = new ColorRange(set, group.Index, group.Length);
                        ranges.Add(range);
                    }
                }
            }

            ranges.Sort();
            Dump(ranges, "Before", line);

            // create intermediate null range
            // note: overlapped match will create undetermined results...
            int lastCovered = 0;
            List<ColorRange> newRanges = new List<ColorRange>();
            foreach (ColorRange range in ranges)
            {
                if (range.StartIndex > lastCovered)
                {
                    Main.Log(" Add range start(last):" + lastCovered + " len:" + (range.StartIndex - lastCovered) + " '" + line.Substring(lastCovered, range.StartIndex - lastCovered) + "'");
                    newRanges.Add(new ColorRange(null, lastCovered, range.StartIndex - lastCovered));
                }

                lastCovered = range.StartIndex + range.Length;
                Main.Log("  last:" + lastCovered);
            }
            if (lastCovered < line.Length)
            {
                // create intermediate null last range
                newRanges.Add(new ColorRange(null, lastCovered, line.Length - lastCovered));
            }

            ranges.AddRange(newRanges);
            
            // sort again (note: could be removed if we change the algorithm and insert intermediate ranges at the same time... not sure it's really useful)
            ranges.Sort();
            Dump(ranges, "After", line);

            foreach (ColorRange range in ranges)
            {
                yield return range;
            }
        }

        private static void Dump(List<ColorRange> ranges, string text, string line)
        {
            Main.Log(text + " Ranges:" + ranges.Count);
            foreach (ColorRange range in ranges)
            {
                Main.Log(" range:" + range + " colorset:" + range.ColorSet + " startIndex:" + range.StartIndex + " length:" + range.Length);
                Main.Log("  text:'" + line.Substring(range.StartIndex, range.Length) + "'");
            }
        }

        public override string ToString()
        {
            return StartIndex + ":" + Length + ":" + ColorSet;
        }

        public override int GetHashCode()
        {
            return Length.GetHashCode() ^ StartIndex.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as ColorRange);
        }

        public bool Equals(ColorRange other)
        {
            if (other == null)
                return false;

            if (Length != other.Length)
                return false;

            if (StartIndex != other.StartIndex)
                return false;

            if (ColorSet == null)
                return other.ColorSet == null;

            return ColorSet.Equals(other.ColorSet);
        }

        int IComparable.CompareTo(object obj)
        {
            return CompareTo(obj as ColorRange);
        }

        public int CompareTo(ColorRange other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            return StartIndex.CompareTo(other.StartIndex);
        }
    }
}
